//*******************************************************************************
// file  uart_avst_adapter.v
//
// parameter
//      DATA_WIDTH    : Avalon-ST data bit width
//      SLAVE_ADDRESS : UART module CSR address
//
// ports
//      clk           : Input clock
//      reset_n       : Input reset
//
//      in_ready      : Avalon-ST Sink(form Tx) Ready
//      in_valid      : Avalon-ST Sink(form Tx) Valid
//      in_data       : Avalon-ST Sink(form Tx) Data
//
//      out_ready     : Avalon-ST Source(to Rx) Ready
//      out_valid     : Avalon-ST Source(to Rx) Valid
//      out_data      : Avalon-ST Source(to Rx) Data
//
//      ctl_address   : Avalon-MM master(to UART CSR) Address
//      ctl_write     : Avalon-MM master(to UART CSR) Write
//      ctl_writedata : Avalon-MM master(to UART CSR) Write Data
//      ctl_read      : Avalon-MM master(to UART CSR) Read
//      ctl_readdata  : Avalon-MM master(to UART CSR) Read Data
//      ctl_waitrequest: Avalon-MM master(to UART CSR) Wait Request
// 
//      csr_address   : Avalon-MM slave(from Nios II) Address
//      csr_write     : Avalon-MM slave(from Nios II) Write
//      csr_writedata : Avalon-MM slave(from Nios II) Write Data
//      csr_read      : Avalon-MM slave(from Nios II) Read
//      csr_readdata  : Avalon-MM slave(from Nios II) Read Data
//
// attention 
// Copyright (C) 2018 MACNICA,Inc. All Rights Reserved.
//   This software is licensed "AS IS". 
//   Please perform use of this software by a user's own responsibility and expense.
//   It cannot guarantee in the maker side about the damage which occurred by the ab-
//   ility not to use or use this software, and all damage that occurred secondarily.
//*******************************************************************************

// synthesis translate_off
`timescale 1ns / 1ps
// synthesis translate_on

`default_nettype none

module uart_avst_adapter
#(
    parameter DATA_WIDTH    = 8,    // Avalon-ST Data bit width
    parameter SLAVE_ADDRESS = 0     // UART module CSR address
)
(
    input   wire                        clk             ,   // clock
    input   wire                        reset_n         ,   // reaset_n
    // IN:Avalon-ST Sink
    output  wire                        in_ready        ,   // Sink ready
    input   wire                        in_valid        ,   // Sink valid
    input   wire    [DATA_WIDTH-1:0]    in_data         ,   // Sink data
    // OUT:Avalon-ST Source
    input   wire                        out_ready       ,   // Source ready
    output  wire                        out_valid       ,   // Source valid
    output  wire    [DATA_WIDTH-1:0]    out_data        ,   // Source data
    // CTL:Avalon-MM Master
    output  reg     [31:0]              ctl_address     ,   // Control port address
    output  reg                         ctl_write       ,   // Control port write
    output  reg     [15:0]              ctl_writedata   ,   // Control port writedata
    output  reg                         ctl_read        ,   // Control port read
    input   wire    [15:0]              ctl_readdata    ,   // Control port readdata
    input   wire                        ctl_waitrequest ,   // Control port wait request
    // CSR:Avalon-MM Slave
    input   wire    [1:0]               csr_address     ,   // Register address
    input   wire                        csr_write       ,   // Register write
    input   wire    [31:0]              csr_writedata   ,   // Register writedata
    input   wire                        csr_read        ,   // Register write
    output  reg     [31:0]              csr_readdata        // Register writedata
);

    // Local parameters.
    localparam INTERVAL     = 32'd100;  // clk / 100

    localparam REG_RX_DATA  = 32'h0 + SLAVE_ADDRESS;
    localparam REG_TX_DATA  = 32'h4 + SLAVE_ADDRESS;
    localparam REG_STATUS   = 32'h8 + SLAVE_ADDRESS;

    localparam ST_IDLE      = 4'd0;     // Idle
    localparam ST_RD_REG    = 4'd1;     // Read status register
    localparam ST_RD_CMP    = 4'd2;     // Read status register is completed

    localparam ST_TX_REG    = 4'd3;     // Receive transmit data from avalon-ST sink
    localparam ST_TX_WAIT   = 4'd4;     // Write transmit data register
    localparam ST_TX_CMP    = 4'd5;     // Wait write latency

    localparam ST_RX_REG    = 4'd6;     // Read receive data register
    localparam ST_RX_WAIT   = 4'd7;     // Wait read latency
    localparam ST_RX_CMP    = 4'd8;     // Transmit receive data to avalon-ST source

    reg     [1:0]               r_csr_state;
    reg     [31:0]              r_csr_timeout;
    reg     [31:0]              r_timer;
    reg     [3:0]               r_state;
    reg                         r_rrdy;
    reg     [DATA_WIDTH-1:0]    r_tx_data;
    reg     [DATA_WIDTH-1:0]    r_rx_data;
    reg     [31:0]              r_tx_count;
    reg     [31:0]              r_rx_count;

    wire                        w_tx_enable;
    wire                        w_rx_enable;
    wire    [15:0]              w_tx_data;
    wire    [DATA_WIDTH-1:0]    w_rx_data;

    assign  in_ready = (r_state == ST_TX_REG)? 1'b1 : 1'b0;
    assign  out_valid = ((r_state == ST_RX_CMP) & out_ready)? 1'b1 : 1'b0;
    assign  out_data = r_rx_data;
    assign  w_tx_enable = r_csr_state[0];
    assign  w_rx_enable = r_csr_state[1];

    generate begin
    if (DATA_WIDTH > 8) begin
        assign w_tx_data = {r_tx_data[7:0],    r_tx_data[15:8]};    // Write transmit data(9-bit) support Little endian
        assign w_rx_data = {ctl_readdata[7:0], ctl_readdata[15:8]}; // Latch receive data(9-bit) support Little endian
    end else begin
        assign w_tx_data = {{(16 - DATA_WIDTH){1'b0}}, r_tx_data};  // Write transmit data(7,8-bit)
        assign w_rx_data = ctl_readdata[DATA_WIDTH-1:0];            // Latch receive data(7,8-bit)
    end
    end endgenerate

    // CSR(Avalon-MM Slave)
    always @(posedge clk or negedge reset_n) begin
        if (!reset_n) begin
            r_csr_state   <= 2'b00;
            r_csr_timeout <= INTERVAL;
        end else begin
            if(csr_write) begin
                r_csr_state   <= (csr_address == 2'd0)? csr_writedata[1:0] : r_csr_state;   // Status/Control reg'
                r_csr_timeout <= (csr_address == 2'd1)? csr_writedata      : r_csr_timeout; // Interval reg'
            end
            if(csr_read) begin
                csr_readdata <= (csr_address == 2'd0)? {30'b0, r_csr_state} :   // Status/Control reg'
                                (csr_address == 2'd1)? r_csr_timeout :          // Interval reg'
                                (csr_address == 2'd2)? r_tx_count :             // Transmit counter
                                (csr_address == 2'd3)? r_rx_count : 32'b0;      // Receive counter
            end
        end
    end

    // Interval counter
    always @(posedge clk or negedge reset_n) begin
        if (!reset_n) begin
            r_timer  <= 32'd0;
        end else begin
            if(r_timer >= r_csr_timeout) begin
                r_timer  <= 32'd0;
            end else begin
                r_timer <= r_timer + 32'd1;
            end
        end
    end

    // State machine
    always @(posedge clk or negedge reset_n) begin
        if (!reset_n) begin
            r_state         <= ST_IDLE;
            ctl_address     <= REG_STATUS;
            ctl_read        <= 1'b0;
            ctl_write       <= 1'b0;
            ctl_writedata   <= 16'h0000;
            r_rrdy          <= 1'b0;
            r_rx_data       <= {(DATA_WIDTH){1'b0}};
            r_tx_data       <= {(DATA_WIDTH){1'b0}};
        end else begin
            // Clear Tx/Rx counted if disable at Rx or Tx
            r_tx_count <= (w_tx_enable)? r_tx_count : 32'b0; 
            r_rx_count <= (w_rx_enable)? r_rx_count : 32'b0; 
            // The processing execution is canceled.
            if(~w_tx_enable & ~w_rx_enable) begin
                r_state <= ST_IDLE;
            end else
            case(r_state)
                // Idle[0]
                ST_IDLE : begin
                    r_state <= (r_timer == 32'd0)? ST_RD_REG : ST_IDLE; // Wait time-out
                end

                //=============================================================
                // The status register of a communication module is read. 
                //=============================================================
                // Read status register[1]
                ST_RD_REG : begin
                    if(!ctl_waitrequest) begin      // waitreq:Lo?
                        ctl_address <= REG_STATUS;  // STATUS reg address
                        ctl_read <= 1'b1;           // read=Hi
                        r_state <= ST_RD_CMP;       // Next state
                    end
                end

                // Wait read latency[2]
                ST_RD_CMP : begin
                    if(!ctl_waitrequest) begin      // waitreq:Lo?
                        ctl_read <= 1'b0;           // read=Lo
                        r_rrdy  <= (w_rx_enable & ctl_readdata[7])? 1'b1 : 1'b0;    // Latch STATUS reg[rrdy:bit-7]
                        r_state <= (w_tx_enable & ctl_readdata[6])? ST_TX_REG :     // Tx enable & trdy:Hi?
                                   (w_rx_enable & ctl_readdata[7])? ST_RX_REG :     // Rx Enable & rrdy:Hi?
                                                                    ST_IDLE;        // if both are Lo return to IDLE
                    end
                end

                //=============================================================
                // Transmit process
                //=============================================================
                // Receive transmit data from avalon-ST sink[4]
                ST_TX_REG : begin
                    if(in_valid) begin              // sink valid enable
                        r_tx_data <= in_data;       // Latch data
                        r_state <= ST_TX_WAIT;      // Next state
                    end else begin
                        r_state <= (r_rrdy)? ST_RX_REG : ST_IDLE; // rredy=Hi? Receive process / Lo IDLE
                    end
                end

                // Write transmit data register[5]
                ST_TX_WAIT : begin
                    if(!ctl_waitrequest) begin      // waitreq:Lo?
                        ctl_address <= REG_TX_DATA; // Transmit Data reg address
                        ctl_write <= 1'b1;          // write=Hi
                        ctl_writedata <= w_tx_data; // Write transmit data
                        r_state <= ST_TX_CMP;       // Next state
                    end
                end

                // Wait write latency[7]
                ST_TX_CMP : begin
                    if(!ctl_waitrequest) begin      // waitreq:Low?
                        ctl_write <= 1'b0;          // write=Lo
                        r_tx_count <= r_tx_count + 32'd1;           // Count tx data
                        r_state <= (r_rrdy)? ST_RX_REG : ST_IDLE;   // rredy=Hi? Receive process / Lo IDLE
                   end
                end

                //=============================================================
                // Receive process
                //=============================================================
                // Read receive data register[8]
                ST_RX_REG : begin
                    if(!ctl_waitrequest) begin      // waitreq:Low?
                        ctl_address <= REG_RX_DATA; // Receive Data reg address
                        ctl_read <= 1'b1;           // read=Hi
                        r_state <= ST_RX_WAIT;      // Next state
                    end
                end

                // Wait read latency[9]
                ST_RX_WAIT : begin
                    if(!ctl_waitrequest) begin      // waitreq:Low?
                        ctl_read <= 1'b0;           // read=Lo
                        r_rx_data <= w_rx_data;     // Latch receive data
                        r_state <= ST_RX_CMP;       // Next state
                    end
                end

                // Transmit receive data to avalon-ST source[11]
                ST_RX_CMP : begin
                    if(out_ready) begin             // source ready:Hi?
                        r_rx_count <= r_rx_count + 32'd1;// Count rx data
                        r_state <= ST_IDLE;         // Return to IDLE
                    end
                end

                default : begin
                    r_state <= ST_IDLE;
                end
            endcase
        end
    end

endmodule

`default_nettype wire
